@@ -1,4 +1,5 @@
 import logging
+import random
 from concurrent.futures import ProcessPoolExecutor, as_completed
 from datetime import date, timedelta
 
@@ -76,14 +77,11 @@ class DataLoader:
     
     def sync_regions(
             self,
-            check_passed: bool
     ) -> bool:
         """
         Ensure `regions` in the company DB matches the general masterdata.
         """
 
-        DataLoader.check_check_passed(check_passed=check_passed)
-
         # Open DBs
         general_db = SQLAlchemyDatabase.from_kedro(credentials_name="postgres_general_data")
         company_db = SQLAlchemyDatabase.from_kedro(credentials_name="postgres")
@@ -152,7 +150,6 @@ class DataLoader:
     ) -> int:
         """Writes the company metadata to both general and company databases and returns the company ID."""
         DataLoader.check_check_passed(check_passed=check_passed)
-        print("regions synced?", regions_synced)
         if not regions_synced:
             raise ValueError("wait for regions to be synced")
 
@@ -483,7 +480,7 @@ class DataLoader:
 
     def write_skus(
         self,
-        time_sku_data: pl.DataFrame,
+        sales_data_map: pl.DataFrame,
         store_ids: pl.DataFrame,
         product_ids: pl.DataFrame,
         company_id: int,
@@ -496,28 +493,28 @@ class DataLoader:
         db = SQLAlchemyDatabase.from_kedro()
 
         # Get unique store-product pairs
-        time_sku_data = (
-            time_sku_data.select(["store", "product"])
+        sales_data_map = (
+            sales_data_map.select(["store", "product"])
             .unique()
             .with_columns(company_id=pl.lit(company_id))
             .join(store_ids.rename({"name": "store"}), on=["store", "company_id"], how="left")
             .join(product_ids.rename({"name": "product"}), on=["product", "company_id"], how="left")
         )
-        time_sku_combinations = time_sku_data.select(["store_id", "product_id", "company_id"])
+        skus = sales_data_map.select(["store_id", "product_id", "company_id"])
 
         # Write to SKU table
         result = db.handle_insertion_multi_line(
             model_class=SkuTable,
-            data=time_sku_combinations,
+            data=skus,
             mode=self.insertion_mode,
             returning_id=True,
         )
 
-        time_sku_data = time_sku_data.join(result, on=["store_id", "product_id", "company_id"], how="left")
+        sales_data_map = sales_data_map.join(result, on=["store_id", "product_id", "company_id"], how="left")
 
         logger.info("Writing SKUs to db complete, got %d sku IDs back", len(result))
 
-        return time_sku_data  # this is sku_ids in the next function
+        return sales_data_map  # this is sku_ids in the next function
 
     def write_time_sku_data(
         self,
@@ -1246,17 +1243,15 @@ class DataLoader:
 
         db = SQLAlchemyDatabase.from_kedro()
 
-        # Get all dates from the database
-        dates_df = get_all_dates(db)
-
         # Get unique combinations of region, country, level to look up region IDs
         region_lookup = time_region_feature_map.select(["region", "country", "level"]).unique()
+        time_region_feature_map = time_region_feature_map.with_columns(company_id=pl.lit(company_id))
 
         with db.get_session(read_only=True) as session:
             # Query region IDs from database
             country_abbrevs = region_lookup.select("country").unique().to_series().to_list()
             country_query = (
-                session.query(Regions.id, Regions.abbreviation)
+                session.query(Regions.region_id, Regions.abbreviation)
                 .filter(Regions.abbreviation.in_(country_abbrevs), Regions.type == "country")
                 .all()
             )
@@ -1277,7 +1272,7 @@ class DataLoader:
             ]
 
             region_results = (
-                session.query(Regions.id, Regions.abbreviation, Regions.type, Regions.country)
+                session.query(Regions.region_id, Regions.abbreviation, Regions.type, Regions.country)
                 .filter(or_(*region_conditions))
                 .all()
             )
@@ -1302,7 +1297,7 @@ class DataLoader:
 
         feature_type = "time_region"
         merge_col_names = ["region", "country", "level"]  # Match on region, country, level
-        select_col_names = ["date_id", "region_id", "feature_id", "value"]  # Include date_id for time features
+        select_col_names = ["date", "region_id", "feature_id", "value"]  # Include date_id for time features
         sql_table_class = TimeRegionFeatures
         sql_table_class_text = TimeRegionFeaturesText
 
@@ -1510,7 +1505,12 @@ def _copy_upsert_month_worker(
                 all_columns=cols,
             )
             rows_to_upsert = _table_rowcount(conn, staging_table)
-            logger.info("Month %s: got %d rows to upsert after deduplication", month_key, rows_to_upsert)
+            logger.info(
+                "Month %s: found %d identical rows to skip, %d rows to upsert after deduplication",
+                month_key,
+                skipped_rows,
+                rows_to_upsert,
+            )
 
         if rows_to_upsert == 0:
             conn.execute(text(f"DROP TABLE IF EXISTS {staging_table}"))
@@ -1523,11 +1523,11 @@ def _copy_upsert_month_worker(
     return row_count
 
 
-def _copy_direct_quarter_worker(
+def _copy_direct_batch_worker(
     *,
     schema: str,
     table: str,
-    quarter_key: str,
+    batch_key: str,
     payload: bytes | str,
     is_binary: bool,
     row_count: int,
@@ -1567,81 +1567,6 @@ def _copy_direct_quarter_worker(
     return row_count
 
 
-# =========================
-# Chunk comparison utilities for TimescaleDB
-# =========================
-def _get_affected_chunks(db, table_name: str, min_date, max_date) -> pl.DataFrame:
-    """Get TimescaleDB chunk information for the given time range."""
-    with db.engine.begin() as conn:
-        query = text(f"""
-            SELECT
-                chunk_schema,
-                chunk_name,
-                range_start::date as chunk_start,
-                range_end::date as chunk_end
-            FROM timescaledb_information.chunks
-            WHERE hypertable_name = '{table_name}'
-              AND range_start <= CAST(:max_date AS timestamp)
-              AND range_end >= CAST(:min_date AS timestamp)
-            ORDER BY range_start
-        """)
-
-        result = conn.execute(query, {
-            "min_date": str(min_date),
-            "max_date": str(max_date)
-        }).fetchall()
-
-        if not result:
-            return pl.DataFrame(schema={
-                "chunk_schema": pl.Utf8,
-                "chunk_name": pl.Utf8,
-                "chunk_start": pl.Date,
-                "chunk_end": pl.Date
-            })
-
-        return pl.DataFrame([
-            {
-                "chunk_schema": row[0],
-                "chunk_name": row[1],
-                "chunk_start": row[2],
-                "chunk_end": row[3]
-            }
-            for row in result
-        ])
-
-
-def _get_chunk_data(db, model_cls, chunk_start, chunk_end, time_column: str) -> pl.DataFrame:
-    """Read compressed data from a specific chunk time range."""
-    table_metadata = _get_table_metadata(model_cls)
-    schema, table_name, _, _, _, table_columns, _ = table_metadata
-
-    with db.engine.begin() as conn:
-        table = model_cls.__table__
-        time_column_obj = getattr(table.c, time_column)
-        selected_columns = [getattr(table.c, column_name) for column_name in table_columns]
-
-        statement = (
-            select(*selected_columns)
-            .where(time_column_obj >= bindparam("chunk_start"))
-            .where(time_column_obj < bindparam("chunk_end"))
-            .order_by(time_column_obj)
-        )
-
-        params = {}
-        for key, raw_value in {"chunk_start": chunk_start, "chunk_end": chunk_end}.items():
-            if hasattr(raw_value, "item"):
-                raw_value = raw_value.item()
-            params[key] = raw_value
-
-        rows = conn.execute(statement, params).mappings().all()
-
-        if not rows:
-            return pl.DataFrame(schema=dict.fromkeys(table_columns, pl.Utf8))
-
-        data = [dict(row) for row in rows]
-        df = pl.DataFrame(data)
-        # Ensure column order matches table definition for deterministic comparisons
-        return df.select([col for col in table_columns if col in df.columns])
 
 
 def _quote_ident(name: str) -> str:
@@ -1742,6 +1667,7 @@ def write_time_features(
 
     # -------- EMPTY TABLE: direct COPY into hypertable, quarterly batching --------
     if table_is_empty:
+        logger.warning("trying out new strategy")
         if effective_strategy is TimeDataStrategy.IGNORE:
             logger.info(
                 "Skipping initial load for %s.%s because strategy is IGNORE.",
@@ -1761,21 +1687,6 @@ def write_time_features(
             _pause_compression(conn, f"{schema}.{table_name}", new_after="100 years")
             _decompress_overlapping_chunks(conn, table_name, min_date, max_date)
 
-        # Build quarterly key: "YYYY-Qn"
-        qdf = df.with_columns(
-            [
-                pl.col(time_column).dt.year().alias("_year"),
-                pl.col(time_column).dt.quarter().alias("_q"),
-            ]
-        ).with_columns(
-            pl.concat_str([pl.col("_year").cast(pl.Utf8), pl.lit("-Q"), pl.col("_q").cast(pl.Utf8)]).alias(
-                "_quarter_key"
-            )
-        )
-
-        # Sort within each quarter for locality (ts, space)
-        sort_cols = [c for c in (time_column, space_column) if c in cols]
-
         # Decide on binary capability:
         # - If underlying driver is psycopg (v3), we'll try binary COPY.
         # - If psycopg2, fall back to CSV (we'd need to hand-pack the binary protocol otherwise).
@@ -1788,35 +1699,49 @@ def write_time_features(
         else:
             logger.info("psycopg v3 not detected; falling back to CSV COPY for initial load.")
 
-        # Prepare quarterly payloads
-        logger.info("Preparing quarterly batches for direct COPY...")
-        quarterly_batches: list[tuple[str, bytes | str, bool, int]] = []  # (qkey, payload, is_binary, rows)
-        for _, subdf in qdf.group_by("_quarter_key", maintain_order=True):
-            subdf = subdf.drop(["_year", "_q"]).sort(sort_cols)
+        # Prepare balanced payloads that span the whole time range while keeping each series together
+        logger.info("Preparing balanced batches for direct COPY...")
 
-            if can_binary:
-                # Try to use psycopg v3's binary COPY writer format via pandas-like pipe:
-                # We don't have a turnkey binary encoder for Polars; CSV is our portable fallback.
-                # To still leverage binary when possible, we use psycopg v3 text encoding only if a
-                # simple adapter isn't available. If you have a custom binary packer, plug it here.
-                # For now we fall back to CSV even with psycopg v3 (copy_expert FORMAT binary requires
-                # packed tuples). Keep the switch for future upgrade.
-                buf = BytesIO()
-                subdf.drop("_quarter_key").write_csv(buf)  # fallback CSV
-                payload = buf.getvalue().decode("utf-8")
-                quarterly_batches.append(
-                    (subdf.select(pl.col("_quarter_key")).unique().item(), payload, False, subdf.height)
-                )
-            else:
-                # CSV payload
-                buf = BytesIO()
-                subdf.drop("_quarter_key").write_csv(buf)
-                payload = buf.getvalue().decode("utf-8")
-                quarterly_batches.append(
-                    (subdf.select(pl.col("_quarter_key")).unique().item(), payload, False, subdf.height)
-                )
+        # Ensure deterministic pseudo-random shuffling of space groups
+        shuffled_series: list[pl.DataFrame]
+        if space_column and space_column in cols:
+            grouped = df.sort([space_column, time_column]).partition_by(space_column, as_dict=True, maintain_order=True)
+            group_items = list(grouped.items())
+            random.Random(0).shuffle(group_items)
+            shuffled_series = [part.sort(time_column) for _, part in group_items]
+        else:
+            shuffled_series = [df.sort(time_column)]
 
-        logger.info("Prepared %d quarterly batches for direct COPY.", len(quarterly_batches))
+        batch_row_limit = 1_000_000
+        copy_batches: list[tuple[str, bytes | str, bool, int]] = []
+        current_parts: list[pl.DataFrame] = []
+        current_rows = 0
+
+        def flush_batch(batch_id: int, parts: list[pl.DataFrame]) -> None:
+            if not parts:
+                return
+            batch_df = pl.concat(parts, how="vertical") if len(parts) > 1 else parts[0]
+            buf = BytesIO()
+            batch_df.write_csv(buf)
+            payload = buf.getvalue().decode("utf-8")
+            copy_batches.append((f"batch_{batch_id:05d}", payload, False, batch_df.height))
+
+        batch_idx = 0
+        for series_df in shuffled_series:
+            rows = series_df.height
+            # Always keep a series intact; start a new batch if it would exceed the row limit
+            if current_rows and current_rows + rows > batch_row_limit:
+                flush_batch(batch_idx, current_parts)
+                batch_idx += 1
+                current_parts = []
+                current_rows = 0
+
+            current_parts.append(series_df)
+            current_rows += rows
+
+        flush_batch(batch_idx, current_parts)
+
+        logger.info("Prepared %d COPY batches for direct load (batch size ≈ %d rows).", len(copy_batches), batch_row_limit)
         total_rows = df.height
         with tqdm(
             total=total_rows, desc=f"Writing {table_name} (fast-load, direct)", unit="rows", dynamic_ncols=True
@@ -1825,25 +1750,25 @@ def write_time_features(
                 with ProcessPoolExecutor(max_workers=max_workers) as ex:
                     futs = [
                         ex.submit(
-                            _copy_direct_quarter_worker,
+                            _copy_direct_batch_worker,
                             schema=schema,
                             table=table_name,
-                            quarter_key=qkey,
+                            batch_key=batch_id,
                             payload=payload,
                             is_binary=is_bin,
                             row_count=rows,
                             cols=cols,
                         )
-                        for (qkey, payload, is_bin, rows) in quarterly_batches
+                        for (batch_id, payload, is_bin, rows) in copy_batches
                     ]
                     for f in as_completed(futs):
                         pbar.update(f.result())
             else:
-                for qkey, payload, is_bin, rows in quarterly_batches:
-                    processed = _copy_direct_quarter_worker(
+                for batch_id, payload, is_bin, rows in copy_batches:
+                    processed = _copy_direct_batch_worker(
                         schema=schema,
                         table=table_name,
-                        quarter_key=qkey,
+                        batch_key=batch_id,
                         payload=payload,
                         is_binary=is_bin,
                         row_count=rows,
@@ -1981,3 +1906,83 @@ def write_time_features(
                         )
 
                 pbar.update(row_count)
+
+
+
+
+
+# =========================
+# Chunk comparison utilities for TimescaleDB
+# =========================
+def _get_affected_chunks(db, table_name: str, min_date, max_date) -> pl.DataFrame:
+    """Get TimescaleDB chunk information for the given time range."""
+    with db.engine.begin() as conn:
+        query = text(f"""
+            SELECT
+                chunk_schema,
+                chunk_name,
+                range_start::date as chunk_start,
+                range_end::date as chunk_end
+            FROM timescaledb_information.chunks
+            WHERE hypertable_name = '{table_name}'
+              AND range_start <= CAST(:max_date AS timestamp)
+              AND range_end >= CAST(:min_date AS timestamp)
+            ORDER BY range_start
+        """)
+
+        result = conn.execute(query, {
+            "min_date": str(min_date),
+            "max_date": str(max_date)
+        }).fetchall()
+
+        if not result:
+            return pl.DataFrame(schema={
+                "chunk_schema": pl.Utf8,
+                "chunk_name": pl.Utf8,
+                "chunk_start": pl.Date,
+                "chunk_end": pl.Date
+            })
+
+        return pl.DataFrame([
+            {
+                "chunk_schema": row[0],
+                "chunk_name": row[1],
+                "chunk_start": row[2],
+                "chunk_end": row[3]
+            }
+            for row in result
+        ])
+
+
+def _get_chunk_data(db, model_cls, chunk_start, chunk_end, time_column: str) -> pl.DataFrame:
+    """Read compressed data from a specific chunk time range."""
+    table_metadata = _get_table_metadata(model_cls)
+    schema, table_name, _, _, _, table_columns, _ = table_metadata
+
+    with db.engine.begin() as conn:
+        table = model_cls.__table__
+        time_column_obj = getattr(table.c, time_column)
+        selected_columns = [getattr(table.c, column_name) for column_name in table_columns]
+
+        statement = (
+            select(*selected_columns)
+            .where(time_column_obj >= bindparam("chunk_start"))
+            .where(time_column_obj < bindparam("chunk_end"))
+            .order_by(time_column_obj)
+        )
+
+        params = {}
+        for key, raw_value in {"chunk_start": chunk_start, "chunk_end": chunk_end}.items():
+            if hasattr(raw_value, "item"):
+                raw_value = raw_value.item()
+            params[key] = raw_value
+
+        rows = conn.execute(statement, params).mappings().all()
+
+        if not rows:
+            return pl.DataFrame(schema=dict.fromkeys(table_columns, pl.Utf8))
+
+        data = [dict(row) for row in rows]
+        df = pl.DataFrame(data)
+        # Ensure column order matches table definition for deterministic comparisons
+        return df.select([col for col in table_columns if col in df.columns])
\ No newline at end of file
